// x86_cfgCreationOracle.C

#include "x86_cfgCreationOracle.h"
#include "simplePath.h"

static bool
handleConditionalBranchInstrTakenCase(bbid_t bb_id,
                                      kptr_t destAddr,
				      const simplePath &pathFromEntryBB,
				         // doesn't include this
                                      const kmem */*km*/,
                                      parseOracle *theParseOracle) 
{
   // Returns true iff we split ourselves, in which case the caller should
   // process the fall-through case, if any, on behalf of getsucc(0) instead 
   // of this.

   // If destAddr isn't in any bb, then recursively create it.
   // Else, if destAddr is the start of a bb, then use it.
   // Else, if it's in the middle of an existing bb, then split it (being 
   // careful when splitting ourselves).

   // Note: we can be assured that this branch is intraprocedural; whoever 
   // called this routine was certain of this before calling.  Therefore, 
   // the following asserts may be applied:
   function_t *fn = theParseOracle->getMutableFn();
   assert(fn->getOrigCode().enclosesAddr(destAddr));
   assert(fn->getBasicBlockFromId(bb_id)->numSuccessors() == 0);

   bbid_t succ_bb_id = fn->searchForBB(destAddr);
   if (succ_bb_id == (bbid_t)-1) {
      // destAddr doesn't lie within an existing basic block.  Recurse.
      basicblock_t *newbb = basicblock_t::getBasicBlock(destAddr, bb_id);
      succ_bb_id = fn->addBlock_new(newbb);

      fn->addSuccessorToBlockByIdNoDup(bb_id, succ_bb_id);
         // won't be a dup since it was just created

      fn->parse(succ_bb_id,
		simplePath(pathFromEntryBB, bb_id), // pathFromEntryBB + bb_id
		theParseOracle);

      return false;
   }

   if (fn->getBasicBlockFromId(succ_bb_id)->getStartAddr() == destAddr) {
      // destAddr is the beginning of an existing bb; use it.
      fn->addSuccessorToBlockByIdNoDup(bb_id, succ_bb_id); 
         // won't be a dup since numSuccessors() was 0
      if (!fn->addPredecessorToBlockByIdUnlessDup(succ_bb_id, bb_id))
	 assert(false);

      return false;
   }

   assert(fn->getBasicBlockFromId(bb_id)->numSuccessors() == 0);

   // destAddr lies in the middle of an existing bb.  Split that basic block
   // in two.  Be careful when splitting ourselves, it's a tricky case
   bbid_t bb_to_split_id = succ_bb_id;
   const bool splitSelf = (bb_to_split_id == bb_id);
   succ_bb_id = fn->splitBasicBlock(bb_to_split_id, destAddr);

   // NOTE: we don't bother to update pathFromEntryBB since it's only needed 
   // if we're gonna recurse ... which doesn't happen here.

   // Finally, do what we came here to do:
   if (!splitSelf) {
      // If we didn't split ourselves, then we should have no successors
      assert(fn->getBasicBlockFromId(bb_id)->numSuccessors() == 0);
      fn->addSuccessorToBlockByIdNoDup(bb_id, succ_bb_id);
         // won't be a dup since numSuccessors() was 0
      if (!fn->addPredecessorToBlockByIdUnlessDup(succ_bb_id, bb_id))
	 assert(false);
   }
   else {
      // If we split ourselves, then we should now have 1 successor: the lower
      // half of the split block
      const basicblock_t *bb = fn->getBasicBlockFromId(bb_id);
      assert(bb->numSuccessors() == 1);

      const bbid_t succ0bb_id = bb->getsucc(0);
      const basicblock_t *succ0bb = fn->getBasicBlockFromId(succ0bb_id);
      
      assert(succ0bb->numPredecessors() == 1);
      assert(succ0bb->getpred(0) == bb_id);

      // Add a successor representing the taken branch. Since we have split 
      // ourselves, the branch should be added to the successors of getsucc(0);
      if (fn->addSuccessorToBlockByIdUnlessDup(succ0bb_id, succ0bb_id))
         if (!fn->addPredecessorToBlockByIdUnlessDup(succ0bb_id, succ0bb_id))
            assert(false);
   }

   // NOTE: If splitSelf, then we want to process the fall-through case on 
   // behalf of 'succ', instead of curr bb.
   return splitSelf;
}

static void
handleConditionalBranchInstrFallThroughCase(bbid_t bb_id,
                                            kptr_t fallThroughInsnAddr,
					    const simplePath &pathFromEntryBB,
					      // curr bb has not yet been added
                                            const kmem */*km*/,
                                            parseOracle *theParseOracle,
                                            bbParsingStats &theBBParsingStats)
{
   // If the dest block is nextKnownFnStartAddr, then do nothing (much).
   // Else, if the dest block lies at the start of an existing bb, then use it.
   // Else, if it lies in the middle on an existing bb, then assert fail (never
   // happens with the fall-through case).

   function_t *fn = theParseOracle->getMutableFn();
   assert(fn->getBasicBlockFromId(bb_id)->numSuccessors() < 2);

   if (!fn->containsAddr(fallThroughInsnAddr)) {
      // This check is new   03/2000 --ari
      // The fall-through case stumbles onto another function, noted by adding
      // "ExitBBId" as a successor instead of parsing the fall-thru block.
      // if (verbose_fnParse) {
      //    cout << "NOTE: Function " << fn->getPrimaryName().c_str()
      //         << " (" << addr2hex(fn->getEntryAddr()) << ")"
      //         << " has a conditional branch that falls through to "
      //         << addr2hex(fallThroughInsnAddr) 
      //         << ", the start of another function" << endl;
      // }

      fn->addSuccessorToBlockByIdNoDup(bb_id, basicblock_t::xExitBBId);
         // stronger assertion!  12/2000

      fn->markBlockAsHavingInterProcFallThru(bb_id);

      theBBParsingStats.numCondBranchFallThrusThatDoInterProceduralStumble++;
      
      return;
   }

   bbid_t succ_id = fn->searchForBB(fallThroughInsnAddr);

   if (succ_id == (bbid_t)-1) {
      // fall-thru addr isn't within any existing basic block, so recursively
      // create it.
      basicblock_t *newbb = basicblock_t::getBasicBlock(fallThroughInsnAddr, bb_id);
      succ_id = fn->addBlock_new(newbb);

      fn->addSuccessorToBlockByIdNoDup(bb_id, succ_id);
         // won't be a duplicate if it didn't exist anywhere until now

      fn->parse(succ_id,
		simplePath(pathFromEntryBB, bb_id), // pathFromEntryBB + bb_id
		theParseOracle);
   }
   else {
      const basicblock_t *fallThruBlock = fn->getBasicBlockFromId(succ_id);
   
      if (fallThruBlock->getStartAddr() == fallThroughInsnAddr) {
         // Dest block lies at the start of an existing bb; use it.
         // Don't allow duplicates in 'successors':
         if (fn->getBasicBlockFromId(bb_id)->numSuccessors() == 0 ||
             fn->getBasicBlockFromId(bb_id)->getsucc(0) != succ_id) { 
	    // disallow duplicates
            fn->addSuccessorToBlockByIdNoDup(bb_id, succ_id);
            if (!fn->addPredecessorToBlockByIdUnlessDup(succ_id, bb_id))
               assert(false);
         }
      }
      else {
         // dest block lies within an existing bb without being the start of 
	 // a bb; can't happen with the fall-through case.  
	 // Proof: Assume fall through address is in the middle of an existing
	 // basic block.  This means that there are insns before it in that 
	 // basic block, because one does not fall through without a branch 
	 // instruction preceding it.  But that's strange, since we've
         // just been working on the instructions before the fall-through 
	 // case; and they're in the current bb, which won't yet show up in 
	 // a searchForBB() yet.
         cout << "Sorry, fall thru to " << addr2hex(fallThroughInsnAddr)
              << " lies within, but not at the beginning of, a basic block "
              << " whose startAddr is " << addr2hex(fallThruBlock->getStartAddr())
              << endl;
         assert(false);
      }
   }

   assert(fn->getBasicBlockFromId(bb_id)->numSuccessors() > 0);
}

void x86_cfgCreationOracle::
processCallRecursion(bbid_t bb_id, kptr_t calleeAddr,
                     instr_t *callInstr,
                     kptr_t instrAddr,
                     bool /*delaySlotInSameBlock*/,
                     instr_t */*dsInsn*/,
                     const simplePath &pathFromEntryBB
                        // doesn't include this block yet
                     ) 
{
   fn.bumpUpBlockSizeBy1Insn(bb_id, callInstr);

   // Recursively parse the destination of the call (will probably do 
   // nothing, since presumably, the destination is the start of the fn, 
   // an already-parsed block).
   const bool splitSelf =
      handleConditionalBranchInstrTakenCase(bb_id,
                                            calleeAddr,
                                            pathFromEntryBB,
                                               // doesn't include this block yet
                                            km,
                                            this);

   // Recursively parse the fall-thru of the call
   
   bbid_t fallThruOnBehalfOf = splitSelf ? 
      fn.getBasicBlockFromId(bb_id)->getsucc(0) : bb_id;
   
   handleConditionalBranchInstrFallThroughCase(fallThruOnBehalfOf,
                                               instrAddr + callInstr->getNumBytes(), 
                                                  // fallthru addr
                                               splitSelf ?
                                                  simplePath(pathFromEntryBB, bb_id) :
                                                     // pathFromEntryBB + bb_id
                                                  pathFromEntryBB,
                                               km,
                                               this,
                                               theBBParsingStats);

   theBBParsingStats.numCallRecursion++;
}

void x86_cfgCreationOracle::
processInterproceduralBranch(bbid_t bb_id, kptr_t destAddr,
                             instr_t */*i*/,
                             const controlFlowInfo_t &/*cfi*/,
                             bool branchAlways,
                             kptr_t /*instrAddr*/,
                             bool /*delaySlotInThisBlock*/,
                             instr_t */*delayInstr*/) 
{
//     if (verbose_fnParse) {
//        cout << "NOTE: inter-procedural branch detected" << endl;
//        cout << "(dest addr is " << addr2hex(destAddr) << endl;
//        cout << "startAddr=" << addr2hex(fn.getEntryAddr()) << endl;
//        cout << "The branch insn is at address " << addr2hex(instrAddr) << ")"
//             << endl;
//     }

   // Add "interProcBranchExitBBId" as a successor to this basic block,
   // and add to interProcJumpsAndCalls[].

   // Note that, like a normal "intra-procedural" branch, it's possible that the
   // fall-thru case could be an inter-procedural stumble.  But that's not our
   // concern in this method.

   fn.addSuccessorToBlockByIdNoDup(bb_id, basicblock_t::interProcBranchExitBBId);
      
   assert(instr_t::instructionHasValidAlignment(destAddr));
   interProcJumpsAndCalls[destAddr] = true; // val is dummy; key is what matters

   if (branchAlways)
      theBBParsingStats.numInterProcBranchAlways++;
   else {
      theBBParsingStats.numInterProcCondBranches++;
   }
}
   
void x86_cfgCreationOracle::
processCondBranch(bbid_t bb_id,
		  bool /* branchAlways */, 
		  bool /* branchNever */,
		  instr_t *i,
		  controlFlowInfo_t *cfi,
		  kptr_t instrAddr,
		  bool /* hasDelaySlotInsn */,
		  bool /* delaySlotInThisBlock */,
		  instr_t * /* delaySlotInsn */,
		  kptr_t fallThroughAddr,
		  kptr_t destAddr,
		  const simplePath &pathFromEntryBB)
{
   fn.bumpUpBlockSizeBy1Insn(bb_id, i);
   assert(fn.getBasicBlockFromId(bb_id)->numSuccessors() == 0);

   bool definitelySplit = false;

   // Handle the taken case.  It may or may not be interprocedural.
   if(!fn.containsAddr(destAddr)) {
      // interprocedural branch
      this->processInterproceduralBranch(bb_id,
					 destAddr, i, *cfi,
					 false, // branchAlways
					 instrAddr,
					 false, // delaySlotInThisBlock
					 NULL); // delaySlotInsn*
      // No need to pass around "path" information, since all this routine
      // does is add the appropriate exit edge.
   }
   else {
      // intraprocedural branch
      theBBParsingStats.numIntraProcCondBranches++;
      definitelySplit = handleConditionalBranchInstrTakenCase(bb_id,
							      destAddr,
							      pathFromEntryBB,
							      km,
							      this);
   }

   // Recurse on the fall-through case, if any.  If curr bb has been split 
   // anytime during processing of the taken case, then the fall-through case 
   // must be processed on behalf of bottom half
   
   bbid_t parentIdOfFallThruBlock = bb_id;
   if(definitelySplit) {
     // bottom half bbid will be the one that contains the branch
     parentIdOfFallThruBlock = fn.searchForBB(instrAddr);
   }
   
   assert(parentIdOfFallThruBlock != (bbid_t)-1);
         
   handleConditionalBranchInstrFallThroughCase(
          parentIdOfFallThruBlock,
	  fallThroughAddr,
	  parentIdOfFallThruBlock != bb_id ?
                 simplePath(pathFromEntryBB, bb_id) : pathFromEntryBB,
	  km,
	  this,
	  theBBParsingStats);
   
}

void x86_cfgCreationOracle::
processNonTailOptimizedTrueCall(bbid_t bb_id, 
                                instr_t *callInstr,
                                kptr_t instrAddr,
                                kptr_t destAddr,
                                bool /*delaySlotInThisBlock*/,
                                instr_t */*delayInstr*/) 
{
   fn.bumpUpBlockSizeBy1Insn(bb_id, callInstr);

   assert(instr_t::instructionHasValidAlignment(destAddr));
   interProcJumpsAndCalls[destAddr] = true;

   theBBParsingStats.numTrueCalls_notTail++;
}
   
void x86_cfgCreationOracle::
processNonTailOptimizedJmplCall(bbid_t bb_id, 
                                instr_t *callInstr,
                                kptr_t instrAddr,
                                const bool delaySlotInSameBlock,
                                instr_t *delaySlotInsn
                                ) 
{
   assert(!"x86_cfgCreationOracle::processNonTailOptimizedJmplCall() should never be called\n");
}

void x86_cfgCreationOracle::
processIndirectTailCall(bbid_t bb_id, 
                        instr_t *jmpInsn,
                        kptr_t insnAddr,
                        bool delaySlotInSameBlock,
                        instr_t *dsInsn) const 
{
   assert(!"x86_cfgCreationOracle::processIndirectTailCall() nyi\n");
}

void x86_cfgCreationOracle::
processIntraproceduralJmpToFixedAddr(bbid_t bb_id,
				     instr_t *instr,
				     kptr_t jmpInsnAddr,
				     bool /*delaySlotInSameBlock*/,
				     instr_t */*dsInsn*/,
				     kptr_t destAddr,
				     const simplePathSlice_t &/*jmpRegSlice*/,
				     const simplePath &pathFromEntryBB)
{
   //  if (verbose_fnParse)
   //     cout << "jump to a constant address: " << addr2hex(destAddr) << endl;

   fn.bumpUpBlockSizeBy1Insn(bb_id, instr);

   assert(fn.getBasicBlockFromId(bb_id)->numSuccessors() == 0);
      
   (void)handleConditionalBranchInstrTakenCase(bb_id, destAddr,
                                               pathFromEntryBB,
                                               km,
                                               this);
      // void the result: we don't care if we split ourself (this basic block),
      // since we're not going to process any fall-through case.

   assert(fn.getBasicBlockFromId(bb_id)->numSuccessors() == 1);

   theBBParsingStats.numJumpToRegFixedAddr_IntraProc++;
}

void x86_cfgCreationOracle::
processInterproceduralJmpToFixedAddr(bbid_t bb_id,
				     instr_t *instr,
				     bool /*delaySlotInSameBlock*/,
				     instr_t *dsInsn,
				     kptr_t destAddr,
				     const simplePath &/*pathFromEntryBB*/)
{
   //  if (verbose_fnParse)
   //     cout << "jump to a constant address: " << addr2hex(destAddr) << endl;

   fn.bumpUpBlockSizeBy1Insn(bb_id, instr);

   assert(fn.getBasicBlockFromId(bb_id)->numSuccessors() == 0);
   fn.addSuccessorToBlockByIdNoDup(bb_id, basicblock_t::xExitBBId);

   interProcJumpsAndCalls[destAddr] = true; // value is a dummy; key is what matters

   theBBParsingStats.numJumpToRegFixedAddr_InterProc++;
}

// ---------- Non-Loading Backwards Jump Table: ----------
void x86_cfgCreationOracle::
processNonLoadingBackwardsJumpTable(bbid_t bb_id,
                                    kptr_t jumpTableStartAddr,
                                    kptr_t jumpTableEndAddr,
                                    instr_t *jmpInsn,
                                    bool delaySlotInThisBlock,
                                    instr_t *delaySlotInsn,
                                    const simplePathSlice_t &justPastEndOfJmpTableSlice,
                                    const simplePath &) 
{
   assert(!"x86_cfgCreationOracle::processNonLoadingBackwardsJumpTable() nyi\n");
}

// ---------- Offset Jump Table 32: ----------
void x86_cfgCreationOracle::
processOffsetJumpTable32WithTentativeBounds(bbid_t bb_id,
                                            instr_t *instr,
                                            kptr_t jmpInsnAddr,
                                            bool delaySlotInThisBlock,
                                            instr_t *delaySlotInsn,
                                            kptr_t jumpTableStartAddr,
                                            const pdvector<kptr_t> &,
                                               // jump table block addrs
                                               // tentative!! May be too many!!
                                            const simplePathSlice_t &theSlice,
                                            const simplePath &pathFromEntryBB) 
{
   assert(!"x86_cfgCreationOracle::processOffsetJumpTable32WithTentativeBounds() nyi\n");
}
