// countedVEvents.C
#include <string.h>
#include "util/h/rope-utils.h"
#include "util/h/out_streams.h"
#include "countedVEvents.h"
#include "metricVirtualizer.h"
#include "simpleMetricFocus.h"
#include "vtimerCountedPrimitive.h"
#include "dataHeap.h"
#include "pidInSetPredicate.h"
#include "stacklist.h"
#include "vtimer.h"
#include "kapi.h"

extern kapi_manager kmgr;
extern metricVirtualizer *theGlobalMetricVirtualizer;

// Initialize the timer data in the kernel memory
void countedVEvents::initialize_timer(kptr_t kernelAddr,
				      unsigned elemsPerVector,
				      unsigned bytesPerStride,
				      kptr_t stopRoutineKernelAddr,
				      kptr_t restartRoutineKernelAddr) const
{
    char vtimer[sizeof_vtimer];
    
    // Prepare the vtimer entry for writing
    memset(vtimer, 0, sizeof_vtimer);

#ifndef sparc_sun_solaris2_7
    // If this is a wall timer, mark it as so
    if(includeEventsSwitchedOut()) {
       char *wall_timer = &vtimer[wall_timer_offset];
       *(unsigned *)wall_timer = 1;
    }
#endif

    // Addresses can not be copied as-is: the client byte order may not
    // match that of the kernel
    kmgr.to_kernel_byte_order(&stopRoutineKernelAddr, sizeof(kptr_t));
    kmgr.to_kernel_byte_order(&restartRoutineKernelAddr, sizeof(kptr_t));
    memcpy(vtimer + stop_routine_offset,
	   &stopRoutineKernelAddr, sizeof(kptr_t));
    memcpy(vtimer + restart_routine_offset,
	   &restartRoutineKernelAddr, sizeof(kptr_t));

    kptr_t addr = kernelAddr;
    for (unsigned i=0; i<elemsPerVector; i++) {
	// Do the writing itself
	kmgr.memcopy(vtimer, addr, sizeof_vtimer);
	addr += bytesPerStride;
    }
}

bool countedVEvents::canInstantiateFocus(const focus &theFocus) const {
   if (theFocus.isRoot() || theFocus.isModule())
      return false;
   else {
      // Function, basic block, or even instruction.
      // We'll return 'success' as long as the fn has been successfully parsed
      // into basic blocks.
       kapi_function kFunc;
       if (kmgr.findFunctionByAddress(theFocus.getFnAddr(), &kFunc) < 0) {
	   assert(false);
       }
       return !kFunc.isUnparsed();
   }
}

// What is does: emits code to stop a vtimer of this metric (if needed), and
// also appends the 8 byte {vtimeraddr, auxdata [rcount]} entity to
// vtimerIter iff stopping. vtimerStart is passed in to aid detection of
// overflow.
kapi_snippet countedVEvents::generateMetricSpecificVStop(
    const kapi_snippet &vtimer_addr_base,
    const kapi_snippet &/*vtimer_start*/,
    const kapi_snippet &vtimer_iter) const
{
    // We need to set rcount to auxdata and start -- to new %pic/%tick
    // and (maybe) update total

    // Get a pointer to the current rcount/start
    vtimerCountedStopPrimitive dummy;
    kapi_arith_expr cpuOffset(dummy.computeCpuOffset());
    kapi_arith_expr start_addr(kapi_plus, vtimer_addr_base, cpuOffset);

    kapi_int64_variable total_field(start_addr); // assume total_field_offset==0
    kapi_int64_variable rcount_and_start(start_field_offset, start_addr);

    // Phase 1: save the timer in the list to restart it on switch-in

    // mask-off the start field
    kapi_arith_expr shifted_rcount(kapi_shift_left, 
				   kapi_arith_expr(kapi_shift_right,
						   rcount_and_start,
						   kapi_const_expr(48)),
				   kapi_const_expr(48));

    // Read new %pic/%tick and truncate it to 48 bits
    kapi_arith_expr timer_at_stop(kapi_shift_right, 
				  kapi_arith_expr(kapi_shift_left, 
						  rawCtrExpr,
						  kapi_const_expr(16)),
				  kapi_const_expr(16));
    
    kapi_arith_expr value_to_save(kapi_bit_or, shifted_rcount,
				  timer_at_stop);

    // Save this vtimer's address and value
    kapi_arith_expr save_addr(kapi_assign, 
			      kapi_int_variable(saved_timer_address_offset,
						vtimer_iter),
			      vtimer_addr_base);
    kapi_arith_expr save_value(kapi_assign, 
			       kapi_int_variable(saved_timer_auxdata_offset,
						 vtimer_iter),
			       value_to_save);
    
    // Increment the iterator
    kapi_arith_expr inc_iter(kapi_assign, vtimer_iter, 
			     kapi_arith_expr(kapi_plus, vtimer_iter,
					     kapi_const_expr(
						 sizeof_saved_timer)));
    
    // Phase 2: update total with the time accrued since it started

    // mask-off the rcount field, leave start
    kapi_arith_expr old_start(kapi_shift_right, 
			      kapi_arith_expr(kapi_shift_left,
					      rcount_and_start,
					      kapi_const_expr(16)),
			      kapi_const_expr(16));
    // Compute the delta and truncate it to 32 bits. This approach
    // lets us handle %pic wraparound: the delta is computed correctly
    // even if %pic wrapped around, thanks to the wonders of the
    // 2-bit-complement arithmetic. FIXME: it may break delta for
    // full-64-bit registers like %tick -- won't be able to measure
    // long-running functions which delta does not fit in 32 bits.
    kapi_arith_expr delta(kapi_shift_right,
			  kapi_arith_expr(kapi_shift_left,
					  kapi_arith_expr(kapi_minus,
							  timer_at_stop,
							  old_start),
					  kapi_const_expr(32)),
			  kapi_const_expr(32));
			  

    kapi_arith_expr update_total(kapi_assign, total_field,
				 kapi_arith_expr(kapi_plus, total_field, 
						 delta));
    // zero-out rcount/start
    kapi_arith_expr update_rcount_start(kapi_assign, rcount_and_start,
					kapi_const_expr(0));

    // Assemble all the pieces together
    kapi_vector<kapi_snippet> sv;
    sv.push_back(save_addr);
    sv.push_back(save_value);
    sv.push_back(inc_iter);
    sv.push_back(update_total);
    sv.push_back(update_rcount_start);
    kapi_sequence_expr do_stop(sv);

    // Do stop only if rcount is non-zero
    kapi_bool_expr rcount_is_zero(kapi_ne, shifted_rcount, kapi_const_expr(0));
    kapi_if_expr stop_if_needed(rcount_is_zero, do_stop);

    // Emit ret at the end
    kapi_vector<kapi_snippet> update_and_ret;
    update_and_ret.push_back(stop_if_needed);
    update_and_ret.push_back(kapi_ret_expr());
    
    return kapi_sequence_expr(update_and_ret);
}

// The opposite of the above routine; they work hand-in-hand.
kapi_snippet countedVEvents::generateMetricSpecificVRestart(
    const kapi_snippet &vtimer_addr_base,
    const kapi_snippet &vtimer_aux_data) const
{
    // We need to set rcount to auxdata and start -- to new %pic/%tick
    // and (maybe) update total

    // Get a pointer to the current rcount/start
    vtimerCountedStopPrimitive dummy;
    kapi_arith_expr cpuOffset(dummy.computeCpuOffset());
    kapi_arith_expr start_addr(kapi_plus, vtimer_addr_base, cpuOffset);
    kapi_int64_variable total_field(start_addr); // assume total_field_offset==0
    kapi_int64_variable rcount_and_start(start_field_offset, start_addr);

    // Read new %pic/%tick and truncate it to 48 bits
    kapi_arith_expr new_start(kapi_shift_right, 
                              kapi_arith_expr(kapi_shift_left, 
                                              rawCtrExpr,
                                              kapi_const_expr(16)),
                              kapi_const_expr(16));
    // Shift new rcount into place
    kapi_arith_expr new_shifted_rcount(kapi_shift_left, 
				       kapi_arith_expr(kapi_shift_right,
						       vtimer_aux_data,
						       kapi_const_expr(48)),
				       kapi_const_expr(48));
    // Modify rcount/start with OR'ed new_rcount/new_start. No need for cas
    // since we are in the exclusive mode (only one cpu should be trying to
    // activate this thread)
    kapi_arith_expr new_rcount_and_start(kapi_bit_or, new_shifted_rcount,
					 new_start);
    kapi_arith_expr update_start(kapi_assign, rcount_and_start,
				 new_rcount_and_start);
    kapi_vector<kapi_snippet> update_start_total_ret;
    update_start_total_ret.push_back(update_start);

    // We want to include switched-out time for wall timers -- update total
    if (includeEventsSwitchedOut()) {
	// Current situation:
	//   new_rcount_and_start = (rcount, new start)
	//   vtimer_aux_data = (rcount, old stop)
	// Let's compute delta = (0, new start - old stop)
	kapi_arith_expr delta(kapi_minus, new_rcount_and_start,
			      vtimer_aux_data);
	kapi_arith_expr update_total(kapi_assign, total_field,
				     kapi_arith_expr(kapi_plus, total_field,
						     delta));
	update_start_total_ret.push_back(update_total);
    }
    update_start_total_ret.push_back(kapi_ret_expr());
    return kapi_sequence_expr(update_start_total_ret);
}

void countedVEvents::
asyncInstantiate(simpleMetricFocus &smf,
                 cmfInstantiatePostSmfInstantiation &postInstantiateCB) const {
   const focus &theFocus = smf.getFocus();
   
   // Step 0: on a metric-specific basic, disallow certain foci.
   // In fact, we can assert it, since outside code should have checked with
   // a call to canInstantiateFocus() before calling us.
   assert(canInstantiateFocus(theFocus));

   // Step 1: Allocate the (one) vtimer
   extern dataHeap *theGlobalDataHeap;

   kptr_t vtimerKernelAddr = theGlobalDataHeap->allocVT();

   smf.addVtimer(vtimerKernelAddr);

#ifdef sparc_sun_solaris2_7
   assert(vtimerKernelAddr % 8 == 0);
   
   // virtualization step 0: change %pcr to a value that we like
   kapi_hwcounter_set oldPcrVal;
   if (oldPcrVal.readConfig() < 0) {
       assert(false);
   }
   kapi_hwcounter_set newPcrVal = 
       queryPostInstrumentationPcr(oldPcrVal, theFocus);
   if (!newPcrVal.equals(oldPcrVal)) {
       if (newPcrVal.writeConfig() < 0) {
	   assert(false);
       }
   }
#elif defined(i386_unknown_linux2_4)
   // MJB TODO: adjust settings of perfctrs for new metric
#endif
   
   // virtualization step 1/5: call referenceVirtualizationCode() to download
   // the utility routines into the kernel (but not actually splice any code to
   // call them).
   theGlobalMetricVirtualizer->referenceVirtualizationCode();

   // virtualization step 2/5: reference metric-specific how-to-stop and
   // how-to-start routines.
   std::pair<kptr_t, kptr_t> stopAndRestartRoutines =
      theGlobalMetricVirtualizer->referenceVRoutinesForMetric(*this);

   // Since only a single stop/restart pair is returned above, it'll only work when
   // the number of vtimer's is exactly one.  Assert that.
   assert(getNumValues() == 1);

   dout << "countedVEvents: stop @ [kernel] "
        << addr2hex(stopAndRestartRoutines.first)
        << " and restart @ [kernel] "
        << addr2hex(stopAndRestartRoutines.second) << endl;

   // virtualization step 3/5: initialize the vtimer(s)
   initialize_timer(vtimerKernelAddr,
		    theGlobalDataHeap->getElemsPerVectorVT(),
		    theGlobalDataHeap->getBytesPerStrideVT(),
		    stopAndRestartRoutines.first, // how-to-stop-if-nec
		    stopAndRestartRoutines.second); // how-to-restart
		    
   // virtualization step 4/5: call virtualizeInitializedTimer(), which will
   // start virtualizing it forthwith.  Of course, since the timer is initially
   // off, and no instrumentation code exists to turn it on, virtualization code
   // won't do much of anything yet, though the code is getting touched.
   theGlobalMetricVirtualizer->virtualizeInitializedTimer(vtimerKernelAddr);
      // pass the kernel address of this vtimer
   
   // virtualization step 5/5: actually splice in the "normal" start & stop
   // routines where the user requested them...

   // --------------------------------------------------------------------
   // NOTE: The remainder of this function is quite similar to the pic0Time
   // and pic0VTime.  We should try to share code.
   // --------------------------------------------------------------------


   kapi_function kFunc;
   kmgr.findFunctionByAddress(theFocus.getFnAddr(), &kFunc);

   const bool basicBlock = (theFocus.getBBId() != (uint16_t)-1);
   kapi_basic_block kBlock;
   if (basicBlock) {
       if (kFunc.findBasicBlockById(theFocus.getBBId(), &kBlock) < 0) {
	   assert(false);
       }
   }

   pdvector<kptr_t> instrumentationPoints;

   // TO DO!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   // For the moment, putting exit point snippets into the metricFocus' vec of
   // snippets first is ESSENTIAL.  This will force (at least in practice) the exit
   // points to be instrumented first.  THIS IS ESSENTIAL, because if the entry
   // points go in first, the timer will get set to wacky values as start-timer
   // instrumentation is executed but stop-timer instrumentation isn't executed because
   // it hasn't been put in yet (it will be there in a few millisecs but that's not
   // good enough).  This probably only appears in practice after kerninstd and
   // kperfmon were split into 2 processes...the real time between instrumentation
   // is suddenly quite high, and this once-rare bug became ubiquitous.
   // A REAL SOLUTION IS NEEDED TO SYNCH-UP "COOPERATIVE" CODE PATCHES!

   // First, do the exit points to the [function/basic block].

   if (basicBlock) {
       kptr_t exitAddr = kBlock.getExitAddr();
       instrumentationPoints += exitAddr;
   }
   else {
       // Exit points to function
       kapi_vector<kapi_point> fnExitPoints;
       if (kFunc.findExitPoints(&fnExitPoints) < 0) {
	   assert(false);
       }
       assert(fnExitPoints.size() > 0);
       kapi_vector<kapi_point>::const_iterator iter = fnExitPoints.begin();
       for (; iter != fnExitPoints.end(); iter++) {
	   instrumentationPoints += iter->getAddress();
       }
   }

   unsigned loop_count = 0;

   for (pdvector<kptr_t>::const_iterator iter = instrumentationPoints.begin();
        iter != instrumentationPoints.end(); ++iter) {
      const kptr_t launchAddr = *iter;
      assert(launchAddr != 0);

      // derived class must supply makeStopPredicate()
//      primitive *primptr = makeStopPrimitive(vtimerAddrs.first); // kernel address

      primitive *primptr = new vtimerCountedStopPrimitive(vtimerKernelAddr,
                                                          rawCtrExpr);
      
      if (theFocus.getPids().size() == 0) {
         snippet s(snippet::prereturn, *primptr);
         smf.addSnippet(launchAddr, s);
      }
      else {
         snippet s(snippet::prereturn,
                   pidInSetPredicate(theFocus.getPids()),
                   *primptr);
         smf.addSnippet(launchAddr, s);
      }

      delete primptr; // was allocated above

      ++loop_count;
   }

   // Second, do entry points to the [function/basic block]
   instrumentationPoints.clear(); // important, lest operator+= below do wacky things.
   
   if (basicBlock) {
       kptr_t startAddr = kBlock.getStartAddr();
       instrumentationPoints += startAddr;
   }
   else {
      instrumentationPoints += kFunc.getEntryAddr();
   }

   loop_count = 0;
   
   for (pdvector<kptr_t>::const_iterator iter = instrumentationPoints.begin();
        iter != instrumentationPoints.end(); ++iter) {
      const kptr_t launchAddr = *iter;
      assert(launchAddr != 0);

      primitive *primptr = new vtimerCountedStartPrimitive(vtimerKernelAddr,
                                                           rawCtrExpr);

      if (theFocus.getPids().size() == 0) {
         snippet s(snippet::preinsn, *primptr);
         smf.addSnippet(launchAddr, s);
      }
      else {
         snippet s(snippet::preinsn,
                   pidInSetPredicate(theFocus.getPids()),
                   *primptr);
         smf.addSnippet(launchAddr, s);
      }

      delete primptr; // was allocated with new
      
      ++loop_count;
   }
   
   // As requested:
   postInstantiateCB.smfInstantiationHasCompleted();
}

static pdvector<kptr_t>
simpleMetricFocus2VTimerKernelAddrs(const simpleMetricFocus &smf) {
   const pdvector<kptr_t> &result = smf.getvtimerKernelAddrs();

   const countedVEvents *theCountedVEvents = dynamic_cast<const countedVEvents*>(&smf.getSimpleMetric());
   assert(theCountedVEvents);
   
   assert(result.size() == theCountedVEvents->getNumValues());

   return result;
}

void countedVEvents::
cleanUpAfterUnsplice(const simpleMetricFocus &smf) const {
   // We have much to do when a metricFocus becomes uninstantiated.
   // unvirtualization steps:
   // 1/5: unsplice the "conventional" start & stop code.  We assume that this
   //      has already been done here.
   // 2/5: call unVirtualizeVTimer() to remove from all_vtimers[] and, if appropriate,
   //      unsplice calls to the cswitchout & in handlers.
   // 3/5: deallocate the vtimer at will
   // 4/5: call dereferenceVRoutinesForMetric(), which may
   //      un-download these how-to-stop and how-to-restart routines from the kernel.
   // 5/5: call dereferenceVirtualizationCode(), which may un-download the
   //      cswitchout and in handlers from the kernel.

   const pdvector<kptr_t> vtimerKernelAddrs = simpleMetricFocus2VTimerKernelAddrs(smf);
   assert(vtimerKernelAddrs.size() == getNumValues());
   assert(getNumValues() == 1);
      // I believe that this is all that'll work
   
   for (pdvector<kptr_t>::const_iterator iter = vtimerKernelAddrs.begin();
        iter != vtimerKernelAddrs.end(); ++iter)
      theGlobalMetricVirtualizer->unVirtualizeVTimer(*iter);

   theGlobalMetricVirtualizer->dereferenceVRoutinesForMetric(*this);

   theGlobalMetricVirtualizer->dereferenceVirtualizationCode();
}
